import * as Mqtt from 'mqtt'
import * as Mongo from 'mongodb'
import * as Stream from 'stream'

var MQTT_BROKER_URL : String = 'mqtt://localhost'
var MONGO_DATABASE_URL : String = 'mongodb://localhost/test'


class MqttStream extends Stream.Readable {
  _read() {
  }
}



class MqttMongo {
  private mqttClient : Mqtt.Client
  private dbHandle : Mongo.Db;
  private dbReady : boolean = false;
  private msgCnt : number = 0;
  private mqttStream : MqttStream

  constructor() {
    this.mqttClient = Mqtt.connect(options['broker'])
    this.mqttClient.on('offline', () => { console.log("mqtt client is offline") })
    this.mqttClient.on('reconnect', () => { console.log("mqtt client is reconnecting") })
    this.mqttClient.on('close', () => { console.log("mqtt connection closed") })

    Mongo.MongoClient.connect(options['database'], 
      { 
        'server': {
          'reconnectTries': 10000,
          'reconnectInterval': 1000,
          'socketOptions': {
            'autoReconnect': false
          }
        }  
      })
      .then(
        (dbHandle: Mongo.Db) => { 
          this.dbHandle = dbHandle
          this.dbReady = true; 
          this.dbHandle.on('error', () => { console.log("error on database") })
          this.dbHandle.on('reconnect', () => { 
            console.log("reconnect on database") 
            this.dbReady = true
          })
          this.dbHandle.on('timeout', () => { console.log("timeout on database") })
          this.dbHandle.on('close', () => { 
            console.log("close on database") 
            this.dbReady = false
          })
        },
        (err: String) => {
          console.log("Unable to connect to database: %s", err)
          process.exit(1)
        }
    )

    this.mqttStream = new MqttStream()
    this.mqttStream.on('data', () => {
      if (this.dbReady) {
        var msg
        while ((msg = this.mqttStream.read()) != null) {
          console.log("Something on the stream: %s", msg)
          var coll = this.dbHandle.collection('mqttMongo')
          coll.insertOne({'a': msg})
            .then(
              (r) => { 
                console.log("successfully inserted into database")
              },
              (err) => { 
                console.log("error when trying to insert into database")
              }
            )
        }
      } else {
        console.log("database currently not available, not reading from stream")
      }
    })

  }

  exec(): void {
    this.mqttClient.on('connect', () => {
      console.log("mqtt client connected to broker")
      this.mqttClient.subscribe('MqttMongo/Command')
      this.mqttClient.publish('MqttMongo/Status', 'hello, started up')
    })

    this.mqttClient.on('message', (topic : string, message : string) => {
      this.msgCnt++;
      console.log(`message received ${this.msgCnt}, topic ${topic}, payload ${message}`)

      if (topic == "MqttMongo/Command" && message == "shutdown") {
        this.mqttClient.end()
      } else {
        this.mqttStream.push(`${topic}:${message}`)
        this.mqttStream.emit('data')
      }
    })
  }
}


import options = require('commander')
options
  .version('0.0.1')
  .option('-b, --broker [broker url]', 'Broker URL', MQTT_BROKER_URL)
  .option('-m, --database [database url]', 'MongoDB Database URL', MONGO_DATABASE_URL)
  .parse(process.argv)

const mqttMongo = new MqttMongo()
mqttMongo.exec()

console.log("MqttMongo started")